@@ -37,13 +37,13 @@
  **/
 
 /* Do not internationalize */
-static const char *months[] = {
+static const char *const months[] = {
 	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
 	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
 };
 
 /* Do not internationalize */
-static const char *days[] = {
+static const char *const days[] = {
 	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
 };
 
@@ -54,15 +54,16 @@ static const int days_before[] = {
 GType
 soup_date_get_type (void)
 {
-	static GType type = 0;
+	static volatile gsize type_volatile = 0;
 
-	if (type == 0) {
-		type = g_boxed_type_register_static (
+	if (g_once_init_enter (&type_volatile)) {
+		GType type = g_boxed_type_register_static (
 			g_intern_static_string ("SoupDate"),
-			(GBoxedCopyFunc)soup_date_copy,
-			(GBoxedFreeFunc)soup_date_free);
+			(GBoxedCopyFunc) soup_date_copy,
+			(GBoxedFreeFunc) soup_date_free);
+		g_once_init_leave (&type_volatile, type);
 	}
-	return type;
+	return type_volatile;
 }
 
 /**
@@ -496,6 +497,49 @@ soup_date_to_string (SoupDate *date, SoupDateFormat format)
 	}
 }
 
+/**
+ * soup_date_to_time_t:
+ * @date: a #SoupDate
+ *
+ * Converts @date to a %time_t.
+ *
+ * If @date is not representable as a %time_t, it will be clamped into
+ * range. (In particular, some HTTP cookies have expiration dates
+ * after "Y2.038k" (2038-01-19T03:14:07Z).)
+ *
+ * Return value: @date as a %time_t
+ **/
+time_t
+soup_date_to_time_t (SoupDate *date)
+{
+	time_t tt;
+
+	/* FIXME: offset, etc */
+
+	if (date->year < 1970)
+		return 0;
+
+	/* If the year is later than 2038, we're guaranteed to
+	 * overflow a 32-bit time_t. (If it's exactly 2038, we'll
+	 * *probably* overflow, but only by a little, and it's easiest
+	 * to test that at the end by seeing if the result has turned
+	 * negative.)
+	 */
+	if (sizeof (time_t) == 4 && date->year > 2038)
+		return (time_t)0x7fffffff;
+
+	tt = (date->year - 1970) * 365;
+	tt += (date->year - 1968) / 4;
+	tt += days_before[date->month] + date->day - 1;
+	if (date->year % 4 == 0 && date->month <= 2)
+		tt--;
+	tt = ((((tt * 24) + date->hour) * 60) + date->minute) * 60 + date->second;
+
+	if (sizeof (time_t) == 4 && tt < 0)
+		return (time_t)0x7fffffff;
+	return tt;
+}
+
 /**
  * soup_date_copy:
  * @date: a #SoupDate
