@@ -10,9 +10,35 @@
 #endif
 
 #include "soup-session-sync.h"
-#include "soup-connection.h"
+#include "soup-session-private.h"
+#include "soup-message-private.h"
 #include "soup-misc.h"
 
+/**
+ * SECTION:soup-session-sync
+ * @short_description: Soup session for blocking I/O in multithreaded
+ * programs.
+ *
+ * #SoupSessionSync is an implementation of #SoupSession that uses
+ * synchronous I/O, intended for use in multi-threaded programs.
+ *
+ * You can use #SoupSessionSync from multiple threads concurrently.
+ * Eg, you can send a #SoupMessage in one thread, and then while
+ * waiting for the response, send another #SoupMessage from another
+ * thread. You can also send a message from one thread and then call
+ * soup_session_cancel_message() on it from any other thread (although
+ * you need to be careful to avoid race conditions, where the message
+ * finishes and is then unreffed by the sending thread just before you
+ * cancel it).
+ *
+ * However, the majority of other types and methods in libsoup are not
+ * MT-safe. In particular, you <emphasis>cannot</emphasis> modify or
+ * examine a #SoupMessage while it is being transmitted by
+ * #SoupSessionSync in another thread. Once a message has been handed
+ * off to #SoupSessionSync, it can only be manipulated from its signal
+ * handler callbacks, until I/O is complete.
+ **/
+
 typedef struct {
 	GMutex *lock;
 	GCond *cond;
@@ -20,10 +46,10 @@ typedef struct {
 #define SOUP_SESSION_SYNC_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SESSION_SYNC, SoupSessionSyncPrivate))
 
 static void  queue_message  (SoupSession *session, SoupMessage *msg,
-			     SoupMessageCallbackFn callback,
-			     gpointer user_data);
+			     SoupSessionCallback callback, gpointer user_data);
 static guint send_message   (SoupSession *session, SoupMessage *msg);
-static void  cancel_message (SoupSession *session, SoupMessage *msg);
+static void  cancel_message (SoupSession *session, SoupMessage *msg,
+			     guint status_code);
 
 G_DEFINE_TYPE (SoupSessionSync, soup_session_sync, SOUP_TYPE_SESSION)
 
@@ -102,7 +128,7 @@ soup_session_sync_new_with_options (const char *optname1, ...)
 typedef struct {
 	SoupSession *session;
 	SoupMessage *msg;
-	SoupMessageCallbackFn callback;
+	SoupSessionCallback callback;
 	gpointer user_data;
 } SoupSessionSyncAsyncData;
 
@@ -111,7 +137,7 @@ async_data_free (SoupSessionSyncAsyncData *sad)
 {
 	g_object_unref (sad->session);
 	g_object_unref (sad->msg);
-	g_free (sad);
+	g_slice_free (SoupSessionSyncAsyncData, sad);
 }
 
 static gboolean
@@ -119,7 +145,7 @@ queue_message_callback (gpointer data)
 {
 	SoupSessionSyncAsyncData *sad = data;
 
-	sad->callback (sad->msg, sad->user_data);
+	sad->callback (sad->session, sad->msg, sad->user_data);
 	async_data_free (sad);
 	return FALSE;
 }
@@ -141,11 +167,11 @@ queue_message_thread (gpointer data)
 
 static void
 queue_message (SoupSession *session, SoupMessage *msg,
-	       SoupMessageCallbackFn callback, gpointer user_data)
+	       SoupSessionCallback callback, gpointer user_data)
 {
 	SoupSessionSyncAsyncData *sad;
 
-	sad = g_new (SoupSessionSyncAsyncData, 1);
+	sad = g_slice_new (SoupSessionSyncAsyncData);
 	sad->session = g_object_ref (session);
 	sad->msg = g_object_ref (msg);
 	sad->callback = callback;
@@ -181,7 +207,7 @@ wait_for_connection (SoupSession *session, SoupMessage *msg)
 				goto try_again;
 			else if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 				conn = NULL;
-			else if (msg->status == SOUP_MESSAGE_STATUS_FINISHED) {
+			else if (soup_message_get_io_status (msg) == SOUP_MESSAGE_IO_STATUS_FINISHED) {
 				/* Message was cancelled while we were
 				 * connecting.
 				 */
@@ -201,7 +227,7 @@ wait_for_connection (SoupSession *session, SoupMessage *msg)
 	g_cond_wait (priv->cond, priv->lock);
 
 	/* See if something bad happened */
-	if (msg->status == SOUP_MESSAGE_STATUS_FINISHED) {
+	if (soup_message_get_io_status (msg) == SOUP_MESSAGE_IO_STATUS_FINISHED) {
 		g_mutex_unlock (priv->lock);
 		return NULL;
 	}
@@ -225,17 +251,17 @@ send_message (SoupSession *session, SoupMessage *msg)
 
 		soup_connection_send_request (conn, msg);
 		g_cond_broadcast (priv->cond);
-	} while (msg->status != SOUP_MESSAGE_STATUS_FINISHED);
+	} while (soup_message_get_io_status (msg) != SOUP_MESSAGE_IO_STATUS_FINISHED);
 
 	return msg->status_code;
 }
 
 static void
-cancel_message (SoupSession *session, SoupMessage *msg)
+cancel_message (SoupSession *session, SoupMessage *msg, guint status_code)
 {
 	SoupSessionSyncPrivate *priv = SOUP_SESSION_SYNC_GET_PRIVATE (session);
 
-	SOUP_SESSION_CLASS (soup_session_sync_parent_class)->cancel_message (session, msg);
+	SOUP_SESSION_CLASS (soup_session_sync_parent_class)->cancel_message (session, msg, status_code);
 	g_cond_broadcast (priv->cond);
 }
 
